# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Base client for authentication and communicating with OSDU."""

import logging
from datetime import datetime
from json import loads
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from .base import OsduBaseCredential

logger = logging.getLogger(__name__)


class OsduTokenCredential(OsduBaseCredential):
    """Refresh token based client for connecting with OSDU."""

    __access_token_expire_date = 0
    __access_token = None
    # __id_token = None

    @property
    def client_id(self) -> str:
        """Client id used for authorisation

        Returns:
            str: client id
        """
        return self._client_id

    @property
    def token_endpoint(self) -> str:
        """Token endpoint for refreshing token

        Returns:
            str: token endpoint
        """
        return self._token_endpoint

    @property
    def refresh_token(self) -> str:
        """The current refresh token

        Returns:
            str: refresh token
        """
        return self._refresh_token

    @property
    def client_secret(self) -> str:
        """The currently used client secret

        Returns:
            str: client secret
        """
        return self._client_secret

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        client_id: str,
        token_endpoint: str,
        refresh_token: str,
        client_secret: str,
    ):
        """Setup the new client

        Args:
            client_id (str): client id for connecting
            token_endpoint (str): token endpoint for refreshing token
            refresh_token (str): refresh token
            client_secret (str): client secret
        """
        super().__init__()
        self._client_id = client_id
        self._token_endpoint = token_endpoint
        self._refresh_token = refresh_token
        self._client_secret = client_secret

    def get_token(self, **kwargs) -> str:
        """
        Check expiration date and return access_token.
        """
        if datetime.now().timestamp() > self.__access_token_expire_date:
            self.refresh_access_token()
        return self.__access_token

    def _refresh_access_token(self) -> dict:
        """
        Send refresh token requests to OpenID token endpoint.

        Return dict with keys "access_token", "expires_in", "scope", "token_type", "id_token".
        """
        body = {
            "grant_type": "refresh_token",
            "refresh_token": self._refresh_token,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = urlencode(body).encode("utf8")
        request = Request(url=self._token_endpoint, data=data, headers=headers)
        try:
            with urlopen(request) as response:
                response_body = response.read()
                return loads(response_body)
        except HTTPError as ex:
            code = ex.code
            message = ex.read().decode("utf8")
            logger.error("Refresh token request failed. %s %s", code, message)
            raise

    def refresh_access_token(self) -> dict:
        """Refresh from refresh token.

        Returns:
            dict: Dictionary representing the returned token
        """
        # for i in range(self.retries + 1):
        #     # try several times if there any error
        #     try:
        #         result = self._refresh_access_token()
        #         break
        #     except HTTPError:
        #         if i == self._retries - 1:
        #             # too many errors, raise original exception
        #             raise
        result = self._refresh_access_token()

        if "access_token" in result:
            # self.__id_token = result["id_token"]
            self.__access_token = result["access_token"]
            self.__access_token_expire_date = datetime.now().timestamp() + result["expires_in"]

            # logger.info("Token is refreshed.")
        else:
            print(result.get("error"))
            print(result.get("error_description"))
            print(result.get("correlation_id"))

        return result  # You may need this when reporting a bug



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------


class CredentialUnavailableError(Exception):
    """The credential could not authenticate because required data or state is unavailable."""

    def __init__(self, message=None, **kwargs):
        self.message = str(message)
        super().__init__(kwargs)



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Base credential for authentication with OSDU."""

from abc import ABC, abstractmethod


# pylint: disable=too-few-public-methods
class OsduBaseCredential(ABC):
    """Abstract base credential class for connecting with OSDU.
    It is not intended to use this directly, rather one of it's subclasses.
    """

    @abstractmethod
    def get_token(self, **kwargs) -> str:
        """Get access token, trying to refresh if needed."""



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
# pylint: disable=C0114
from .base import OsduBaseCredential
from .environment import OsduEnvironmentCredential
from .msal_interactive import OsduMsalInteractiveCredential
from .token import OsduTokenCredential

__all__ = [
    "OsduBaseCredential",
    "OsduEnvironmentCredential",
    "OsduTokenCredential",
    "OsduMsalInteractiveCredential",
]



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Search client for working with the OSDU search API."""

from typing import Union

from osdu.client import OsduClient
from osdu.serviceclientbase import ServiceClientBase

VALID_SEARCH_API_VERSIONS = [2]


# pylint: disable=too-few-public-methods
class SearchClient(ServiceClientBase):
    # Dev. notes:
    # inspiration from:
    # https://github.com/Azure/azure-sdk-for-python/blob/3fe8964c8831c9ce91c4a4bc0dadcbc525b74220/sdk/keyvault/azure-keyvault-keys/azure/keyvault/keys/_client.py
    # https://github.com/Azure/azure-sdk-for-python/blob/3fe8964c8831c9ce91c4a4bc0dadcbc525b74220/sdk/keyvault/azure-keyvault-certificates/azure/keyvault/certificates/_shared/client_base.py#L34
    # TO DO Model, or string / dict based API calls!
    # TO DO Async v non async calls
    # TO DO Cursor / paging support.
    """A client for working with the OSDU Search API."""

    def __init__(self, client: OsduClient, service_version: Union[int, str] = "latest"):
        """Setup the SearchClient

        Args:
            client (OsduClient): client to use for connection
            service_version (Union[int, str], optional): service version (3 or 'latest') Defaults to 'latest'.

        Raises:
            ValueError: [description]
        """
        super().__init__(client, "search", VALID_SEARCH_API_VERSIONS, service_version)

    # def query():
    #     pass

    # def query_by_id():
    #     pass

    def is_healthy(self) -> bool:
        """Returns health status of the API

        Returns:
            bool: health status of the API
        """
        response = self._client.get(self.api_url("health/readiness_check"))
        return response.status_code == 200

    def query_all_aggregated(self) -> dict:
        """Returns a list of all kinds including number of records

        Returns:
            dict: containing the result
        """
        request_data = {"kind": "*:*:*:*", "limit": 1, "query": "*", "aggregateBy": "kind"}
        response_json = self._client.post_returning_json(self.api_url("query"), request_data)
        return response_json

    def query(
        self, kind: str = None, identifier: str = None, query: str = None, limit: int = None
    ) -> dict:
        """Query records

        Args:
            kind (str): kind to query for
            identifier (str): id to query for
            query (str): a specific query
            limit (str): limit on number of records to return

        Returns:
            dict: containing the result
        """

        if identifier is not None and query is not None:
            raise ValueError("You can't specify both identifier and query")

        request_data = {}
        if kind is None:
            request_data["kind"] = "*:*:*:*"
        else:
            request_data["kind"] = kind

        if identifier is not None:
            request_data["query"] = f'id:("{identifier}")'

        if query is not None:
            request_data["query"] = query

        if limit is not None:
            request_data["limit"] = limit

        response_json = self._client.post_returning_json(self.api_url("query"), request_data)
        return response_json

    def query_by_id(self, identifier: str, limit: int = None) -> dict:
        """Returns a list of all kinds including number of records

        Args:
            identifier (str): id to query for

        Returns:
            dict: containing the result
        """
        request_data = {
            "kind": "*:*:*:*",
            "query": f'id:("{identifier}")',
        }

        if limit is not None:
            request_data["limit"] = limit

        response_json = self._client.post_returning_json(self.api_url("query"), request_data)
        return response_json

    def query_by_kind(self, kind: str, limit: int = None) -> dict:
        """Returns a list of all records for the given kind

        Args:
            kind (str): kind to query for

        Returns:
            dict: containing the result
        """
        request_data = {"kind": kind}

        if limit is not None:
            request_data["limit"] = limit

        response_json = self._client.post_returning_json(self.api_url("query"), request_data)
        return response_json



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""OSDU SDK package that can be installed using setuptools"""

# For reasons behind /src/ file structure see https://blog.ionelmc.ro/2014/05/25/python-packaging/#the-structure

import os
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
from setuptools import setup, find_packages


def read(fname):
    """Local read helper function for long documentation"""
    osdu_path = dirname(os.path.realpath(__file__))
    return open(join(osdu_path, fname)).read()


version_file = read(os.path.join("src", "osdu", "__init__.py"))
__VERSION__ = re.search(
    r'^__VERSION__\s*=\s*[\'"]([^\'"]*)[\'"]', version_file, re.MULTILINE
).group(1)

setup(
    name="osdu-sdk",
    version=__VERSION__,
    description="OSDU SDK for Python",
    long_description=read("README.rst"),
    url="https://github.com/equinor/osdu-sdk-python",
    author="Equinor ASA",
    author_email="mhew@equinor.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="osdu sdk python",
    python_requires=">=3.8",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    install_requires=["requests", "msal"],
    project_urls={
        "Issue Tracker": "https://github.com/equinor/osdu-sdk-python/issues",
    },
)
